# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

from collections import namedtuple
from copy import deepcopy as copy

import numpy as np
from scipy.sparse import csr_matrix as sparsematrix
from scipy.interpolate import interp1d

import meshio

from .forms import IntegralFormMixed
from .solve import partition, solve
from .math import identity, grad, dot, transpose, eigvals, det, interpolate, norms
from .field import Field

from .doftools import apply, partition as dofpartition, extend as dofextend


def dofresiduals(region, r, rref, dof1=None):

    rx = r
    ry = rref

    ry[ry == 0] = np.nan

    rxy = rx / ry

    rxy[region.mesh.elements_per_node == 1] = rx[region.mesh.elements_per_node == 1]

    if dof1 is None:
        return rxy
    else:
        return rxy.ravel()[dof1]


def newtonrhapson(
    region,
    fields,
    u0ext,
    fun_f,
    fun_A,
    dof1,
    dof0,
    unstack,
    maxiter=20,
    tol=1e-6,
    parallel=True,
):

    # extract fields
    # u, p, J = fields
    dV = region.dV

    # deformation gradient at integration points
    F = identity(grad(fields[0])) + grad(fields[0])

    # PK1 stress and elasticity matrix
    f = fun_f(F, *[interpolate(f) for f in fields[1:]])
    A = fun_A(F, *[interpolate(f) for f in fields[1:]])

    # assembly
    r = IntegralFormMixed(f, fields, dV).assemble(parallel=parallel).toarray()[:, 0]
    K = IntegralFormMixed(A, fields, dV).assemble(parallel=parallel)

    converged = False

    for iteration in range(maxiter):

        system = partition(fields, K, dof1, dof0, r)
        dfields = np.split(solve(*system, u0ext), unstack)

        if np.any(np.isnan(dfields[0])):
            break
        else:
            for field, dfield in zip(fields, dfields):
                field += dfield

        # deformation gradient at integration points
        F = identity(grad(fields[0])) + grad(fields[0])

        # PK1 stress and elasticity matrix
        f = fun_f(F, *[interpolate(f) for f in fields[1:]])

        # residuals and stiffness matrix components
        r = IntegralFormMixed(f, fields, dV).assemble(parallel=parallel).toarray()[:, 0]

        rref = np.linalg.norm(r[dof0])
        uref = np.linalg.norm(fields[0].values.ravel()[dof0])
        if rref == 0:
            rref = 1
        if uref == 0:
            uref = 1
        norm_r = np.linalg.norm(r[dof1[dof1 < len(dfields[0])]]) / rref

        norm_u = np.linalg.norm(dfields[0][dof1[dof1 < len(dfields[0])]]) / uref
        norm_dfields = norms(dfields[1:])

        info_r = f"#{iteration+1:2d}: |r|={norm_r:1.3e} |u|={norm_u:1.3e}"
        info_f = [f"(|δ{2+i}|={norm_f:1.3e})" for i, norm_f in enumerate(norm_dfields)]

        print(" ".join([info_r, *info_f]))

        if norm_r < tol:  # np.array(norm_fields).sum() +
            converged = True
            break
        else:
            # elasticity matrix
            A = fun_A(F, *[interpolate(f) for f in fields[1:]])

            # assembly of stiffness matrix components
            K = IntegralFormMixed(A, fields, dV).assemble(parallel=parallel)

    Result = namedtuple(
        "Result", ["fields", "r", "K", "F", "f", "A", "unstack", "converged"]
    )

    return Result(copy(fields), r, K, F, f, A, unstack, converged)


def incsolve(
    fields,
    region,
    f,
    A,
    bounds,
    move,
    boundary="move",
    filename="out",
    maxiter=8,
    tol=1e-6,
    parallel=True,
):

    res = []

    # dofs to dismiss and to keep
    dof0, dof1, unstack = dofpartition(fields, bounds)
    # solve newton iterations and save result
    for increment, move_t in enumerate(move):

        print(f"\nINCREMENT {increment+1:2d}   (move={move_t:1.3g})")
        # set new value on boundary
        bounds[boundary].value = move_t

        # obtain external displacements for prescribed dofs
        u0ext = apply(fields[0], bounds, dof0)

        Result = newtonrhapson(
            region,
            fields,
            u0ext,
            f,
            A,
            dof1,
            dof0,
            unstack,
            maxiter=maxiter,
            tol=tol,
            parallel=parallel,
        )

        fields = copy(Result.fields)

        if not Result.converged:
            # reset counter for last converged increment and break
            increment = increment - 1
            break
        else:
            # save results and go to next increment
            res.append(Result)
            save(region, *Result, filename=filename)
            # save(region, *Result, filename=filename + f"_{increment+1:d}")
            print("SAVED TO FILE")

    savehistory(region, res, filename=filename)

    return res


def tonodes(values, region, sym=True, mode="tensor"):

    rows = region.mesh.connectivity.T.ravel()
    cols = np.zeros_like(rows)

    ndim = region.mesh.ndim

    if mode == "tensor":
        if sym:
            if ndim == 3:
                ij = [(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)]
            elif ndim == 2:
                ij = [(0, 0), (1, 1), (0, 1)]
        else:
            if ndim == 3:
                ij = [
                    (0, 0),
                    (0, 1),
                    (0, 2),
                    (1, 0),
                    (1, 1),
                    (1, 2),
                    (2, 0),
                    (2, 1),
                    (2, 2),
                ]
            elif ndim == 2:
                ij = [(0, 0), (0, 1), (1, 0), (1, 1)]

        out = Field(region, dim=len(ij)).values

        for a, (i, j) in enumerate(ij):
            out[:, a] = (
                sparsematrix(
                    (values.reshape(ndim, ndim, -1)[i, j], (rows, cols)),
                    shape=(region.mesh.nnodes, 1),
                ).toarray()[:, 0]
                / region.mesh.elements_per_node
            )

    elif mode == "scalar":
        out = sparsematrix(
            (values.ravel(), (rows, cols)), shape=(region.mesh.nnodes, 1)
        ).toarray()[:, 0]
        out = out / region.mesh.elements_per_node

    return out


def save(
    region,
    fields,
    r=None,
    K=None,
    F=None,
    f=None,
    A=None,
    unstack=None,
    converged=True,
    filename="result.vtk",
):

    if unstack is not None:
        u = fields[0]
    else:
        u = fields

    mesh = region.mesh

    point_data = {
        "Displacements": u.values,
    }

    if r is not None:
        if unstack is not None:
            reactionforces = np.split(r, unstack)[0]
        else:
            reactionforces = r

        point_data["ReactionForce"] = reactionforces.reshape(*u.values.shape)

    if f is not None:
        # 1st Piola Kirchhoff stress
        if unstack is not None:
            P = f[0]
        else:
            P = f

        # cauchy stress at integration points
        s = dot(P, transpose(F)) / det(F)
        sp = eigvals(s)

        # shift stresses to nodes and average nodal values
        cauchy = tonodes(s, region=region, sym=True)
        cauchyprinc = [tonodes(sp_i, region=region, mode="scalar") for sp_i in sp]

        point_data["CauchyStress"] = cauchy

        point_data["MaxPrincipalCauchyStress"] = cauchyprinc[2]
        point_data["IntPrincipalCauchyStress"] = cauchyprinc[1]
        point_data["MinPrincipalCauchyStress"] = cauchyprinc[0]

    mesh = meshio.Mesh(
        points=mesh.nodes,
        cells={mesh.etype: mesh.connectivity},  # [:, : mesh.edgenodes]},
        # Optionally provide extra data on points, cells, etc.
        point_data=point_data,
    )

    mesh.write(filename)


def savehistory(region, results, filename="result_history.xdmf"):

    mesh = region.mesh
    points = mesh.nodes
    cells = {mesh.etype: mesh.connectivity}  # [:, : mesh.edgenodes]}

    with meshio.xdmf.TimeSeriesWriter(filename) as writer:
        writer.write_points_cells(points, cells)

        for inc, result in enumerate(results):
            fields, r, K, F, f, A, unstack, converged = result

            if unstack is not None:
                reactionforces = np.split(r, unstack)[0]
            else:
                reactionforces = r

            u = fields[0]

            point_data = {
                "Displacements": u.values,
                "ReactionForce": reactionforces.reshape(*u.values.shape),
            }

            if f is not None:
                # cauchy stress at integration points
                s = dot(f[0], transpose(F)) / det(F)
                sp = eigvals(s)

                # shift stresses to nodes and average nodal values
                cauchy = tonodes(s, region=region, sym=True)
                cauchyprinc = [
                    tonodes(sp_i, region=region, mode="scalar") for sp_i in sp
                ]

                point_data["CauchyStress"] = cauchy

                point_data["MaxPrincipalCauchyStress"] = cauchyprinc[2]
                point_data["IntPrincipalCauchyStress"] = cauchyprinc[1]
                point_data["MinPrincipalCauchyStress"] = cauchyprinc[0]

            writer.write_data(inc, point_data=point_data)


def force(results, boundary):
    return np.array(
        [
            (((np.split(res.r, res.unstack)[0]).reshape(-1, 3))[boundary.nodes]).sum(0)
            for res in results
        ]
    )


def moment(results, boundary, point=np.zeros(3)):

    nodes = results[0].fields[0].region.mesh.nodes
    points = point.reshape(-1, 3)

    indices = np.array([(1, 2), (2, 0), (0, 1)])

    if points.shape[0] == 1:
        points = np.tile(points, (len(results), 1))

    moments = []

    for pt, res in zip(points, results):
        displacements = res.fields[0].values
        d = ((nodes + displacements) - pt)[boundary.nodes]

        force = (np.split(res.r, res.unstack)[0]).reshape(-1, 3)
        f = force[boundary.nodes]

        moments.append([(f[:, i] * d[:, i[::-1]]).sum() for i in indices])

    return np.array(moments)


def curve(x, y):
    if len(y) > 1:
        kind = "linear"
    if len(y) > 2:
        kind = "quadratic"
    if len(y) > 3:
        kind = "cubic"

    f = interp1d(x[: len(y)], y, kind=kind)
    xx = np.linspace(x[0], x[: len(y)][-1])
    return np.array([x[: len(y)], y]), np.array([xx, f(xx)])



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np


class Mesh:
    def __init__(self, nodes, connectivity, etype=None):
        self.nodes = nodes
        self.connectivity = connectivity
        self.etype = etype

        self.update(self.connectivity)

    def update(self, connectivity):
        self.connectivity = connectivity

        # obtain dimensions
        self.nnodes, self.ndim = self.nodes.shape
        self.ndof = self.nodes.size
        self.nelements = self.connectivity.shape[0]

        # get number of elements per node
        nodes_in_conn, self.elements_per_node = np.unique(
            connectivity, return_counts=True
        )

        # check if there are nodes without element connectivity
        if self.nnodes != len(self.elements_per_node):
            self.node_has_element = np.isin(np.arange(self.nnodes), nodes_in_conn)
            # update "epn" ... elements per node
            epn = -np.ones(self.nnodes, dtype=int)
            epn[nodes_in_conn] = self.elements_per_node
            self.elements_per_node = epn

            self.nodes_without_elements = np.arange(self.nnodes)[~self.node_has_element]
            self.nodes_with_elements = np.arange(self.nnodes)[self.node_has_element]
        else:
            self.nodes_without_elements = np.array([], dtype=int)
            self.nodes_with_elements = np.arange(self.nnodes)

    def save(self, filename="mesh.vtk"):
        "Export mesh as VTK file."

        if self.etype is None:
            raise TypeError("Element type missing.")
        else:
            import meshio

        cells = {self.etype: self.connectivity}
        meshio.Mesh(self.nodes, cells).write(filename)


class Cube(Mesh):
    def __init__(self, a=(0, 0, 0), b=(1, 1, 1), n=(2, 2, 2)):
        self.a = a
        self.b = b
        self.n = n

        nodes, connectivity = cube_hexa(a, b, n)
        etype = "hexahedron"

        super().__init__(nodes, connectivity, etype)
        # self.edgenodes = 8


class Rectangle(Mesh):
    def __init__(self, a=(0, 0), b=(1, 1), n=(2, 2)):
        self.a = a
        self.b = b
        self.n = n

        nodes, connectivity = rectangle_quad(a, b, n)
        etype = "quad"

        super().__init__(nodes, connectivity, etype)
        # self.edgenodes = 4


class Line(Mesh):
    def __init__(self, a=0, b=1, n=2):
        self.a = a
        self.b = b
        self.n = n

        nodes, connectivity = line_line(a, b, n)

        etype = "line"

        super().__init__(nodes, connectivity, etype)
        # self.edgenodes = 2


class CubeQuadratic(Mesh):
    def __init__(self, a=(0, 0, 0), b=(1, 1, 1)):
        self.a = a
        self.b = b
        self.n = (3, 3, 3)

        nodes, connectivity = cube_hexa(a, b, n=self.n)
        etype = "hexahedron20"

        super().__init__(nodes, connectivity, etype)
        # self.edgenodes = 8

        self.nodes = self.nodes[
            [0, 2, 8, 6, 18, 20, 26, 24, 1, 5, 7, 3, 19, 23, 25, 21, 9, 11, 17, 15]
        ]
        self.connectivity = np.arange(20).reshape(1, -1)
        self.update(self.connectivity)


class CubeAdvanced(Cube):
    def __init__(
        self,
        n=5,
        L=1,
        B=1,
        H=1,
        dL=0,
        dB=0,
        exponent=4,
        symmetry=(False, False, False),
        L0=0,
        B0=0,
    ):

        a = -np.ones(3)
        symmetry = np.array(symmetry, dtype=bool)
        a[list(symmetry)] = 0
        super().__init__(a, (1, 1, 1), n)

        if L0 > 0 or B0 > 0:
            mask = np.logical_or(self.nodes[:, 0] > L0 / 2, self.nodes[:, 1] > B0 / 2)
            keep = np.arange(self.nnodes)[mask]
            select = np.array(
                [np.all(np.isin(conn, keep)) for conn in self.connectivity]
            )
            self.connectivity = self.connectivity[select]

        z = self.nodes.copy()
        z[:, 0] *= L / 2 * (1 + 2 * dL / L * self.nodes[:, 2] ** exponent)
        z[:, 1] *= B / 2 * (1 + 2 * dB / B * self.nodes[:, 2] ** exponent)
        z[:, 2] *= H / 2
        self.nodes = z
        self.update(self.connectivity)


class CylinderAdvanced(Mesh):
    def __init__(
        self,
        D=10,
        H=1,
        n=(13, 13, 9),
        d=2,
        phi=180,
        dD=1,
        dd=1,
        k=4,
        align=True,
        symmetry=False,
    ):

        R = D / 2
        r = d / 2
        # rm = (R + r) / 2
        dr = R - r

        if symmetry:
            a = 0
        else:
            a = -1

        N, C = expand(line_line(a=a, b=1, n=n[2]), n[0], z=2)
        N[:, 1] -= 1

        bottom = N[:, 1] < 0
        top = N[:, 1] > 0
        Nb = N[bottom]
        Nt = N[top]

        Nb[:, 1] *= 1 + dd / dr * Nb[:, 0] ** k
        Nt[:, 1] *= 1 + dD / dr * Nt[:, 0] ** k

        N[bottom] = Nb
        N[top] = Nt

        N[:, 1] += 1
        N[:, 1] *= dr / 2
        N[:, 1] += r

        N[:, 0] *= H / 2
        N[:, 0] += H / 2

        nodes, connectivity = revolve((N, C), n[1], -phi, axis=0)
        etype = "hexahedron"

        if align:
            nodes, connectivity = rotate(rotate((nodes, connectivity), 90, 1), 90, 2)

        super().__init__(nodes, connectivity, etype)
        # self.edgenodes = 8


class Cylinder(CylinderAdvanced):
    def __init__(self, D=2, H=1, n=(3, 9, 3), phi=360):

        super().__init__(D, H, n, d=0, phi=phi, dD=0, dd=0, k=4, align=True)


# line, rectangle (based on line) and cube (based on rectangle) generators
# ------------------------------------------------------------------------


def line_line(a=0, b=1, n=2):
    "Line generator."
    nodes = np.linspace(a, b, n).reshape(-1, 1)
    connectivity = np.repeat(np.arange(n), 2)[1:-1].reshape(-1, 2)

    return nodes, connectivity


def rectangle_quad(a=(0, 0), b=(1, 1), n=(2, 2)):
    "Rectangle generator."
    dim = 2
    array_like = (tuple, list, np.ndarray)

    # check if number "n" is scalar or no. of nodes per axis (array-like)
    if not isinstance(n, array_like):
        n = np.full(dim, n, dtype=int)

    line = line_line(a[0], b[0], n[0])

    nodes, connectivity = expand(line, n[-1], b[-1] - a[-1])
    nodes[:, -1] += a[-1]

    return nodes, connectivity


def cube_hexa(a=(0, 0, 0), b=(1, 1, 1), n=(2, 2, 2)):
    "Cube generator."
    dim = 3
    array_like = (tuple, list, np.ndarray)

    # check if number "n" is scalar or no. of nodes per axis (array-like)
    if not isinstance(n, array_like):
        n = np.full(dim, n, dtype=int)

    rectangle = rectangle_quad(a[:-1], b[:-1], n[:-1])

    nodes, connectivity = expand(rectangle, n[-1], b[-1] - a[-1])
    nodes[:, -1] += a[-1]

    return nodes, connectivity


def expand(mesh, n=11, z=1):
    "Expand 1d line to 2d quad or 2d quad to 3d hexahedron mesh."

    if isinstance(mesh, Mesh):
        Nodes = mesh.nodes
        Connectivity = mesh.connectivity
        return_mesh = True
    else:
        Nodes, Connectivity = mesh
        return_mesh = False

    Dim = Nodes.shape[1]
    if Dim == 1:
        sl = slice(None, None, -1)
        etype = "quad"
    elif Dim == 2:
        sl = slice(None, None, None)
        etype = "hexahedron"
    else:
        raise ValueError("Expansion of a 3d mesh is not supported.")

    p = np.pad(Nodes, (0, 1))[:-1]
    zeros = np.zeros(Dim)
    nodes = np.vstack([p + np.array([*zeros, h]) for h in np.linspace(0, z, n)])

    c = [Connectivity + len(p) * a for a in np.arange(n)]
    connectivity = np.vstack([np.hstack((a, b[:, sl])) for a, b in zip(c[:-1], c[1:])])

    if return_mesh:
        return Mesh(nodes, connectivity, etype)
    else:
        return nodes, connectivity


def rotation_matrix(alpha_deg, dim=3, axis=0):
    "2d or 3d rotation matrix around specified axis."
    a = np.deg2rad(alpha_deg)
    rotation_matrix = np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])
    if dim == 3:
        # rotation_matrix = np.pad(rotation_matrix, (1, 0))
        # rotation_matrix[0, 0] = 1
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((1, 2)), axis=0)
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((3, 1)), axis=1)
        rotation_matrix[axis, axis] = 1

    return rotation_matrix


def rotate(mesh, angle_deg, axis):
    "Rotate mesh."

    if isinstance(mesh, Mesh):
        Nodes = mesh.nodes
        Connectivity = mesh.connectivity
        Etype = mesh.etype
        return_mesh = True
    else:
        Nodes, Connectivity = mesh
        return_mesh = False

    dim = Nodes.shape[1]
    nodes = (rotation_matrix(angle_deg, dim, axis) @ Nodes.T).T

    if return_mesh:
        return Mesh(nodes, Connectivity, Etype)
    else:
        return nodes, Connectivity


def revolve(mesh, n=11, phi=180, axis=0):
    "Revolve 2d quad to 3d hexahedron mesh."

    if isinstance(mesh, Mesh):
        Nodes = mesh.nodes
        Connectivity = mesh.connectivity
        return_mesh = True
    else:
        Nodes, Connectivity = mesh
        return_mesh = False

    Dim = Nodes.shape[1]
    if Dim == 1:
        sl = slice(None, None, -1)
        etype = "quad"
    elif Dim == 2:
        sl = slice(None, None, None)
        etype = "hexahedron"
    else:
        raise ValueError("Revolution of a 3d mesh is not supported.")

    if abs(phi) > 360:
        raise ValueError("phi must be within |phi| <= 360 degree.")

    p = np.pad(Nodes, (0, 1))[:-1]
    R = rotation_matrix

    nodes = np.vstack(
        [(R(alpha_deg, Dim + 1) @ p.T).T for alpha_deg in np.linspace(0, phi, n)]
    )

    c = [Connectivity + len(p) * a for a in np.arange(n)]

    # if abs(phi) == 360:
    #    nodes = nodes[: -len(p)]
    #    c[-1] = Connectivity

    connectivity = np.vstack([np.hstack((a, b[:, sl])) for a, b in zip(c[:-1], c[1:])])

    nodes, connectivity = sweep((nodes, connectivity), decimals=6)

    if return_mesh:
        return Mesh(nodes, connectivity, etype)
    else:
        return nodes, connectivity


def sweep(mesh, decimals=6):
    "Sweep duplicated nodes and update connectivity."

    if isinstance(mesh, Mesh):
        Nodes = mesh.nodes
        Connectivity = mesh.connectivity
        etype = mesh.etype
        return_mesh = True
    else:
        Nodes, Connectivity = mesh
        return_mesh = False

    nodes, index, inverse, counts = np.unique(
        np.round(Nodes, decimals), True, True, True, axis=0
    )

    original = np.arange(len(Nodes))

    mask = inverse != original
    find = original[mask]
    replace = inverse[mask]

    connectivity = Connectivity.copy()

    for i, j in zip(find, replace):
        connectivity[Connectivity == i] = j

    if return_mesh:
        return Mesh(nodes, connectivity, etype)
    else:
        return nodes, connectivity


def convert(mesh, order=0, calc_nodes=False):
    "Convert mesh to a given order (currently only order=0 supported)."

    if order != 0:
        raise NotImplementedError("Unsupported order conversion.")

    if calc_nodes:
        nodes = np.stack(
            [np.mean(mesh.nodes[conn], axis=0) for conn in mesh.connectivity]
        )
    else:
        nodes = np.zeros((mesh.nelements, mesh.ndim), dtype=int)

    connectivity = np.arange(mesh.nelements).reshape(-1, 1)
    etype = "None"
    return Mesh(nodes, connectivity, etype)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 29 18:29:15 2021

@author: adutz
"""
# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from .math import det, inv


class Region:
    def __init__(self, mesh, element, quadrature):
        self.mesh = mesh
        self.element = element
        self.quadrature = quadrature
        self.connectivity = mesh.connectivity
        self.nodes = mesh.nodes
        self.nnodes = mesh.nnodes
        self.nelements = mesh.nelements

        # array with degrees of freedom
        # h_ap
        # ----
        # basis function "a" evaluated at quadrature point "p"
        self.h = np.array([self.element.basis(p) for p in self.quadrature.points]).T

        # dhdr_aJp
        # --------
        # partial derivative of basis function "a"
        # w.r.t. natural coordinate "J" evaluated at quadrature point "p"
        self.dhdr = np.array(
            [self.element.basisprime(p) for p in self.quadrature.points]
        ).transpose(1, 2, 0)

        if self.element.nbasis > 1:

            # dXdr_IJpe and its inverse drdX_IJpe
            # -----------------------------------
            # geometric gradient as partial derivative of undeformed coordinate "I"
            # w.r.t. natural coordinate "J" evaluated at quadrature point "p"
            # for every element "e"
            dXdr = np.einsum(
                "eaI,aJp->IJpe", self.mesh.nodes[self.connectivity], self.dhdr
            )
            drdX = inv(dXdr)

            # dV_pe = det(dXdr)_pe * w_p
            # determinant of geometric gradient evaluated at quadrature point "p"
            # for every element "e" multiplied by corresponding quadrature weight
            # denoted as "differential volume element"
            self.dV = det(dXdr) * self.quadrature.weights.reshape(-1, 1)

            # dhdX_aJpe
            # ---------
            # partial derivative of basis function "a"
            # w.r.t. undeformed coordinate "J" evaluated at quadrature point "p"
            # for every element "e"
            self.dhdX = np.einsum("aIp,IJpe->aJpe", self.dhdr, drdX)

    def volume(self, detF=1):
        "Calculate element volume for element 'e'."
        return np.einsum("pe->e", detF * self.dV)


# TODO
# class AxiRegion(Region):

#     def __init__(self, mesh, element, quadrature):
#         super.__init__(mesh, element, quadrature)

#         self.R = np.einsum("ea,ap->pe", self.nodes[:,[1,]][self.connectivity], self.h)
#         zeros = np.zeros(self.dhdX.shape[-2:])
#         h = np.tile(self.h.reshape(*self.h.shape,-1), self.nelements)
#         self.dhdX = np.block([[self.dhdX, zeros],
#                               [zeros.T, h/self.R]])



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import pytest
import numpy as np
import felupe as fe


def test_helpers():
    mesh = fe.mesh.Cube()
    dof = np.arange(mesh.ndof).reshape(mesh.nodes.shape)

    u = np.zeros_like(mesh.nodes)

    f0 = lambda x: x == 0
    f1 = lambda x: x == 1
    v = 13.2
    left = fe.doftools.Boundary(dof, mesh, fx=f0, skip=(0, 0, 1), value=v)
    right = fe.doftools.Boundary(dof, mesh, fx=f1, skip=(0, 1, 1))

    bounds = [left, right]

    dof0, dof1 = fe.doftools.partition(dof, bounds)

    uext = fe.doftools.apply(u, dof, bounds, dof0=None)
    u0ext = fe.doftools.apply(u, dof, bounds, dof0=dof0)

    dof01 = np.sort(np.append(dof0, dof1))
    if not np.all(np.equal(dof.ravel(), dof01)):
        raise ValueError("Partitioning of DOF failed.")

    if not np.allclose(uext.ravel()[dof0], u0ext):
        raise ValueError("Application of external values failed.")

    if not np.allclose(uext.ravel()[left.dof], v):
        raise ValueError("Application of external values failed.")

    return mesh, dof, u, uext, left, right, dof0, dof1


if __name__ == "__main__":
    mesh, dof, u, uext, left, right, dof0, dof1 = test_helpers()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import pytest
import numpy as np
import felupe as fe


def test_domain():
    element = fe.element.Hex1()
    mesh = fe.mesh.Cube(n=5)
    quadrature = fe.quadrature.Linear(dim=3)

    domain = fe.domain.Domain(element, mesh, quadrature)

    # elemental volumes and total domain volume
    Ve = domain.volume()
    Vd = Ve.sum()

    if not np.isclose(Vd, 1):
        raise ValueError("Error in Domain Volume calculation.")

    u = domain.zeros()
    x = domain.zeros((1, 7))
    y = domain.fill(value=-12.3, dim=5)
    z = domain.empty(dim=(6, 3, 7))

    if not x.shape == (domain.nnodes, 1, 7):
        raise ValueError("Error in domain.zeros() function.")

    return domain


if __name__ == "__main__":
    domain = test_domain()



from .__about__ import __version__
from . import math
from . import mesh
from . import quadrature
from . import doftools
from . import element
from . import field

# from . import region
from . import constitution
from . import solve
from . import utils

from .region import Region
from .forms import IntegralForm, IntegralFormMixed
from .field import Field
from .doftools import Boundary


__all__ = [
    "__version__",
]



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import pytest
import felupe as fe


def test_mesh():
    mr1 = fe.mesh.Rectangle()
    mr2 = fe.mesh.Rectangle(a=(-1.2, -2), b=(2, 3.1), n=(4, 9))
    mr3 = fe.mesh.Rectangle(n=7)

    mc1 = fe.mesh.Cube()
    mc2 = fe.mesh.Cube(a=(-1, -2, -0.5), b=(2, 3.1, 1), n=(4, 9, 5))
    mc3 = fe.mesh.Cube(n=7)


if __name__ == "__main__":
    test_mesh()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from .math import (
    ddot,
    transpose,
    inv,
    dya,
    cdya,
    cdya_ik,
    cdya_il,
    det,
    identity,
    trace,
)


class LinearElastic:
    def __init__(self, E, nu=0.3):
        self.E = E
        self.nu = nu
        self.mu, self.gamma = self.lame(E, nu)

    def stress(self, strain):
        return 2 * self.mu * strain + self.gamma * trace(strain) * identity(strain)

    def elasticity(self, strain):
        I = identity(strain)
        return self.mu * cdya(I, I) + self.gamma * dya(I, I)

    def lame(self, E, nu):
        mu = E / (2 * (1 + nu))
        gamma = E * nu / ((1 + nu) * (1 - 2 * nu))
        return mu, gamma


class NeoHooke:
    "Nearly-incompressible Neo-Hooke material."

    def __init__(self, mu, bulk):
        self.mu = mu
        self.bulk = bulk

    def P(self, F):
        """Variation of total potential w.r.t displacements
        (1st Piola Kirchhoff stress).

        δ_u(Π_int) = ∫_V ∂ψ/∂F : δF dV

        """

        mu = self.mu
        bulk = self.bulk

        J = det(F)
        iFT = transpose(inv(F, J))

        Pdev = mu * (F - ddot(F, F) / 3 * iFT) * J ** (-2 / 3)
        Pvol = bulk * (J - 1) * J * iFT

        return Pdev + Pvol

    def A(self, F):
        """Linearization w.r.t. displacements of variation of
        total potential energy w.r.t displacements.

        Δ_u(δ_u(Π_int)) = ∫_V δF : ∂²ψ/(∂F∂F) : ΔF dV

        """

        mu = self.mu
        bulk = self.bulk

        J = det(F)
        iFT = transpose(inv(F, J))
        eye = identity(F)

        A4_dev = (
            mu
            * (
                cdya_ik(eye, eye)
                - 2 / 3 * dya(F, iFT)
                - 2 / 3 * dya(iFT, F)
                + 2 / 9 * ddot(F, F) * dya(iFT, iFT)
                + 1 / 3 * ddot(F, F) * cdya_il(iFT, iFT)
            )
            * J ** (-2 / 3)
        )

        p = bulk * (J - 1)
        q = p + bulk * J

        A4_vol = J * (q * dya(iFT, iFT) - p * cdya_il(iFT, iFT))

        return A4_dev + A4_vol



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np


def values(fields):
    if isinstance(fields, tuple):
        return np.concatenate([field.values.ravel() for field in fields])
    else:
        return fields.values.ravel()


def norms(arrays):
    return [np.linalg.norm(arr) for arr in arrays]


def interpolate(A):
    return A.interpolate()


def grad(A):
    return A.grad()


def sym(A):
    return (A + transpose(A)) / 2


def identity(A):
    _, ndim, g, e = A.shape
    return np.tile(np.eye(ndim), (g, e, 1, 1)).transpose([2, 3, 0, 1])


def dya(A, B, mode=2):
    if mode == 2:
        return np.einsum("ij...,kl...->ijkl...", A, B)
    elif mode == 1:
        return np.einsum("i...,j...->ij...", A, B)
    else:
        raise ValueError("unknown mode. (1 or 2)", mode)


def _linalginv(A):
    return np.linalg.inv(A.T).T


def inv(A, detA=None, full_output=False):
    invA = np.zeros_like(A)

    if detA is None:
        detA = det(A)

    if A.shape[0] == 3:
        invA[0, 0] = -A[1, 2] * A[2, 1] + A[1, 1] * A[2, 2]
        invA[1, 0] = A[1, 2] * A[2, 0] - A[1, 0] * A[2, 2]
        invA[2, 0] = -A[1, 1] * A[2, 0] + A[1, 0] * A[2, 1]
        invA[0, 1] = A[0, 2] * A[2, 1] - A[0, 1] * A[2, 2]
        invA[1, 1] = -A[0, 2] * A[2, 0] + A[0, 0] * A[2, 2]
        invA[2, 1] = A[0, 1] * A[2, 0] - A[0, 0] * A[2, 1]
        invA[0, 2] = -A[0, 2] * A[1, 1] + A[0, 1] * A[1, 2]
        invA[1, 2] = A[0, 2] * A[1, 0] - A[0, 0] * A[1, 2]
        invA[2, 2] = -A[0, 1] * A[1, 0] + A[0, 0] * A[1, 1]

    elif A.shape[0] == 2:
        invA[0, 0] = A[1, 1]
        invA[0, 1] = -A[0, 1]
        invA[1, 0] = -A[1, 0]
        invA[1, 1] = A[0, 0]

    if full_output:
        return invA / detA, detA
    else:
        return invA / detA


def _linalgdet(A):
    return np.linalg.det(A.T).T


def det(A):
    if A.shape[0] == 3:
        detA = (
            A[0, 0] * A[1, 1] * A[2, 2]
            + A[0, 1] * A[1, 2] * A[2, 0]
            + A[0, 2] * A[1, 0] * A[2, 1]
            - A[2, 0] * A[1, 1] * A[0, 2]
            - A[2, 1] * A[1, 2] * A[0, 0]
            - A[2, 2] * A[1, 0] * A[0, 1]
        )
    elif A.shape[0] == 2:
        detA = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    return detA


def cof(A):
    return transpose(inv(A), detA=1.0)


def eig(A):
    wA, vA = np.linalg.eig(A.transpose([2, 3, 0, 1]))
    return wA.transpose([2, 0, 1]), vA.transpose([2, 3, 0, 1])


def eigvals(A, shear=False):
    wA = np.linalg.eig(A.transpose([2, 3, 0, 1]))[0].transpose([2, 0, 1])
    if shear:
        ndim = wA.shape[0]
        if ndim == 3:
            ij = [(1, 0), (2, 0), (2, 1)]
        elif ndim == 2:
            ij = [(1, 0)]
        dwA = np.array([wA[i] - wA[j] for i, j in ij])
        return np.vstack((wA, dwA))
    else:
        return wA


def transpose(A):
    return A.transpose([1, 0, 2, 3])


def majortranspose(A):
    return np.einsum("ijkl...->klij...", A)


def trace(A):
    return np.trace(A)


def cdya_ik(A, B):
    return np.einsum("ij...,kl...->ikjl...", A, B)


def cdya_il(A, B):
    return np.einsum("ij...,kl...->ilkj...", A, B)


def cdya(A, B):
    return (cdya_ik(A, B) + cdya_il(A, B)) * 0.5


def dot(A, B):
    if len(A.shape) == len(B.shape):
        return np.einsum("ik...,kj...->ij...", A, B)
    elif len(A.shape) + 2 == len(B.shape):
        return np.einsum("ik...,kj...->ij...", A, B)
    elif len(A.shape) == len(B.shape) + 2:
        return np.einsum("ijkm...,ml...->ijkl...", A, B)
    else:
        raise TypeError("Unknown shape of A and B.")


def ddot(A, B):
    if len(A.shape) == len(B.shape):
        return np.einsum("ij...,ij...->...", A, B)
    elif len(A.shape) + 2 == len(B.shape):
        return np.einsum("ij...,ijkl...->kl...", A, B)
    elif len(A.shape) == len(B.shape) + 2:
        return np.einsum("ijkl...,kl...->ij...", A, B)
    else:
        raise TypeError("Unknown shape of A and B.")


def tovoigt(A):
    B = np.zeros((6, *A.shape[-2:]))
    ij = [(0, 0), (1, 1), (2, 2), (0, 0), (1, 2), (0, 2)]
    for i6, (i, j) in enumerate(ij):
        B[i6, :, :] = A[i, j, :, :]
    return B


def tovoigt2(A):
    B = np.zeros((A.shape[0], 6))
    ij = [(0, 0), (1, 1), (2, 2), (0, 0), (1, 2), (0, 2)]
    for i6, (i, j) in enumerate(ij):
        B[:, i6] = A[:, i, j]
    return B



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
from types import SimpleNamespace


def get_dof0(field, bounds):
    "Extract prescribed degrees of freedom."

    # get mesh from field and obtain field-dimension
    mesh = field.region.mesh
    dim = field.dim

    # check if there are nodes without connected elements in the mesh
    # and add them to the list of prescribed dofs
    # e.g. these are nodes [2,6,7]
    #
    #   ( [[2,2,2], )   [[0,1,2],   [[ 6, 7, 8],
    # 3*(  [6,6,6], ) +  [0,1,2], =  [18,19,20],
    #   (  [7,7,7]] )    [0,1,2]]    [21,22,23]]
    #
    # fixmissing = [6, 7, 8, 18, 19, 29, 21, 22, 23]
    fixmissing = dim * np.tile(mesh.nodes_without_elements, (dim, 1)).T + np.arange(dim)

    # obtain prescribed dofs from boundaries
    dof0_bounds = np.concatenate([b.dof for b in bounds.values()])

    # combine all prescribed dofs and remove repeated itmes if there are any
    return np.unique(np.append(fixmissing.ravel(), dof0_bounds))


def get_dof1(field, bounds, dof0=None):
    "Extract active (non-prescribed) degrees of freedom."

    # obtain all dofs from the field
    dof = field.indices.dof

    # if function argument `dof0` is None call `get_dof0()`
    if dof0 is None:
        dof0 = get_dof0(field, bounds)

    # init a mask for the selection of active dofs
    mask = np.ones_like(dof.ravel(), dtype=bool)

    # set mask items for prescribed dofs (dof0) to False
    mask[dof0] = False

    # make the dof list 1d and mask active dofs
    return dof.ravel()[mask]


# convenient alias names
find = SimpleNamespace()
find.dof0 = get_dof0
find.dof1 = get_dof1


def partition(fields, bounds):
    "Partition dof-list to prescribed and active parts."

    # if a tuple is passed it is assumed that the first
    # field is associated to the boundaries
    if isinstance(fields, tuple):
        f = fields[0]
        extend_dof1 = True
    else:
        f = fields
        extend_dof1 = False

    dof0 = get_dof0(f, bounds)
    dof1 = get_dof1(f, bounds, dof0=dof0)

    # extend active dofs with dofs from additional fields
    if extend_dof1:
        dof0, dof1, offsets = extend(fields, dof0, dof1)
        return dof0, dof1, offsets
    else:
        return dof0, dof1


def extend(fields, dof0, dof1):

    # get sizes of fields and calculate offsets
    fieldsizes = [f.indices.dof.size for f in fields]
    offsets = np.cumsum(fieldsizes)

    # init extended dof0, dof1 arrays
    dof0_xt = dof0.copy()
    dof1_xt = dof1.copy()

    # loop over fields starting from the second ones_like
    for field, offset, fieldsize in zip(fields[1:], offsets[:-1], fieldsizes[1:]):

        # obtain the mesh and the dimension from the current field
        mesh = field.region.mesh
        dim = field.dim

        # check if there are nodes without/with connected elements in the mesh
        # and add them to the list of prescribed/active dofs
        # e.g. these are mesh.nodes_without_elements = [2,6,7]
        #
        #              ( [[2,2,2], )   [[0,1,2],   [[ 6, 7, 8],
        # dof0_add = 3*(  [6,6,6], ) +  [0,1,2], =  [18,19,20],
        #              (  [7,7,7]] )    [0,1,2]]    [21,22,23]]
        #
        dof0_add = (
            offset
            + dim * np.tile(mesh.nodes_without_elements, (dim, 1)).T
            + np.arange(dim)
        )
        dof1_add = (
            offset
            + dim * np.tile(mesh.nodes_with_elements, (dim, 1)).T
            + np.arange(dim)
        )

        dof0_xt = np.append(dof0_xt, dof0_add.ravel())
        dof1_xt = np.append(dof1_xt, dof1_add.ravel())

    return dof0_xt, dof1_xt, offsets[:-1]


def apply(v, bounds, dof0=None):
    """Apply prescribed values for a list of boundaries
    and return all (default) or only the prescribed components
    of the input array 'v' based on the keyword 'dof0'."""

    u = v.values.copy()

    for b in bounds.values():
        u.ravel()[b.dof] = b.value

    if dof0 is None:
        return u
    else:
        # check if dof0 has entries beyond the size of u
        # this is the case for meshes with nodes that are
        # not connected to elements
        u0ext = u.ravel()[dof0[dof0 < u.size]]

        # pad (=extend) u0ext to the full dimension of prescribed dofs
        # and fill padded values with zeros
        u0ext_padded = np.pad(u0ext, (0, len(dof0) - len(u0ext)))
        return u0ext_padded


def symmetry(field, axes=(True, True, True), x=0, y=0, z=0, bounds=None):
    "Create symmetry boundary conditions."

    # obtain the mesh from the field
    # mesh = field.region.mesh

    # convert axes to array and slice by mesh dimension
    enforce = np.array(axes).astype(bool)[: field.dim]

    # create search functions for x,y,z - axes
    fx = lambda v: np.isclose(v, x)
    fy = lambda v: np.isclose(v, y)
    fz = lambda v: np.isclose(v, z)

    # invert boolean identity matrix and use its rows
    # for the skip argument (a symmetry condition on
    # axis "z" fixes all displacements u_z=0 but keeps
    # in-plane displacements active)
    skipax = ~np.eye(3).astype(bool)
    kwarglist = [
        {"fx": fx, "skip": skipax[0][: field.dim]},
        {"fy": fy, "skip": skipax[1][: field.dim]},
        {"fz": fz, "skip": skipax[2][: field.dim]},
    ]

    if bounds is None:
        bounds = {}
    labels = ["symx", "symy", "symz"]

    # loop over symmetry conditions and add them to a new dict
    for a, (symaxis, kwargs) in enumerate(zip(enforce, kwarglist[: field.dim])):
        if symaxis:
            bounds[labels[a]] = Boundary(field, **kwargs)

    return bounds


class Boundary:
    def __init__(
        self,
        field,
        name="default",
        fx=lambda v: v == np.nan,
        fy=lambda v: v == np.nan,
        fz=lambda v: v == np.nan,
        value=0,
        skip=(False, False, False),
    ):

        mesh = field.region.mesh
        dof = field.indices.dof

        self.ndim = field.dim  # mesh.ndim
        self.name = name
        self.value = value
        self.skip = np.array(skip).astype(int)[: self.ndim]
        self.fun = [fx, fy, fz][: self.ndim]

        # apply functions on the nodes per coordinate
        # fx(x), fy(y), fz(z) and create a mask for each coordinate
        mask = [f(x) for f, x in zip(self.fun, mesh.nodes.T)]

        # combine the masks with logical or
        tmp = np.logical_or(mask[0], mask[1])
        if self.ndim == 3 and mesh.nodes.shape[1] == 3:
            mask = np.logical_or(tmp, mask[2])
        else:
            mask = tmp

        # tile the mask
        self.mask = np.tile(mask.reshape(-1, 1), self.ndim)

        # check if some axes should be skipped
        if True not in skip:
            pass
        else:
            # exclude mask from axes which should be skipped
            self.mask[:, np.where(self.skip)[0]] = False

        self.dof = dof[self.mask]
        self.nodes = np.arange(mesh.nnodes)[mask]



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import quadpy


class GaussLegendre:
    def __init__(self, order, dim):
        # integration point weights and coordinates
        self.dim = dim
        self.npoints = (order + 1) ** dim

        if dim == 3:
            scheme = quadpy.c3.product(quadpy.c1.gauss_legendre(order + 1))
        elif dim == 2:
            scheme = quadpy.c2.product(quadpy.c1.gauss_legendre(order + 1))
        elif dim == 1:
            scheme = quadpy.c1.gauss_legendre(order + 1)
        else:
            raise ValueError("Wrong dimension.")

        self.points = scheme.points.T
        self.weights = scheme.weights * 2 ** dim


class Constant(GaussLegendre):
    def __init__(self, dim):
        super().__init__(order=0, dim=dim)


class Linear(GaussLegendre):
    def __init__(self, dim):
        super().__init__(order=1, dim=dim)


class Quadratic(GaussLegendre):
    def __init__(self, dim):
        super().__init__(order=2, dim=dim)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
from scipy.sparse import csr_matrix as sparsematrix
from scipy.sparse import bmat, vstack


class IntegralFormMixed:
    def __init__(self, fun, fields, dV, grad=None):

        self.fun = fun
        self.fields = fields
        self.nfields = len(fields)
        self.dV = dV

        if grad is None:
            self.grad = np.zeros_like(fields, dtype=bool)
            self.grad[0] = True
        else:
            self.grad = grad

        self.forms = []

        if self.nfields == 1:
            raise ValueError("IntegralFormMixed needs at least 2 fields.")

        if len(fun) == self.nfields:
            self.mode = 1
            self.i = np.arange(self.nfields)
            self.j = np.zeros_like(self.i)

            for fun, field, grad_field in zip(self.fun, self.fields, self.grad):
                f = IntegralForm(fun=fun, v=field, dV=self.dV, grad_v=grad_field)
                self.forms.append(f)

        elif len(fun) == np.sum(1 + np.arange(self.nfields)):
            self.mode = 2
            self.i, self.j = np.triu_indices(3)

            for a, (i, j) in enumerate(zip(self.i, self.j)):
                f = IntegralForm(
                    self.fun[a],
                    v=self.fields[i],
                    dV=self.dV,
                    u=self.fields[j],
                    grad_v=self.grad[i],
                    grad_u=self.grad[j],
                )
                self.forms.append(f)
        else:
            raise ValueError("Unknown input format.")

    def assemble(self, values=None, parallel=False, block=True):

        out = []

        if values is None:
            values = [None] * len(self.forms)

        for val, form in zip(values, self.forms):
            out.append(form.assemble(val, parallel))

        if block and self.mode == 2:
            K = np.zeros((self.nfields, self.nfields), dtype=object)
            for a, (i, j) in enumerate(zip(self.i, self.j)):
                K[i, j] = out[a]
                if i != j:
                    K[j, i] = out[a].T

            return bmat(K).tocsr()

        if block and self.mode == 1:
            return vstack(out).tocsr()

        else:
            return out

    def integrate(self, parallel=False):

        out = []
        for form in self.forms:
            out.append(form.integrate(parallel))

        return out


class IntegralForm:
    def __init__(self, fun, v, dV, u=None, grad_v=False, grad_u=False):
        self.fun = fun
        self.dV = dV

        self.v = v
        self.grad_v = grad_v

        self.u = u
        self.grad_u = grad_u

        if not self.u:
            self.indices = self.v.indices.ai
            self.shape = self.v.indices.shape

        else:
            eai = self.v.indices.eai
            ebk = self.u.indices.eai

            eaibk0 = np.stack(
                [np.repeat(ai.ravel(), bk.size) for ai, bk in zip(eai, ebk)]
            )
            eaibk1 = np.stack(
                [np.tile(bk.ravel(), ai.size) for ai, bk in zip(eai, ebk)]
            )

            self.indices = (eaibk0.ravel(), eaibk1.ravel())
            self.shape = (self.v.indices.shape[0], self.u.indices.shape[0])

    def assemble(self, values=None, parallel=False):

        if values is None:
            values = self.integrate(parallel=parallel)

        permute = np.append(len(values.shape) - 1, range(len(values.shape) - 1)).astype(
            int
        )

        out = sparsematrix(
            (values.transpose(permute).ravel(), self.indices), shape=self.shape
        )

        return out

    def integrate(self, parallel=False):
        grad_v, grad_u = self.grad_v, self.grad_u
        v, u = self.v, self.u
        dV = self.dV
        fun = self.fun

        # if parallel:
        #    from numba import jit

        if not grad_v:
            vb = np.tile(
                v.region.h.reshape(*v.region.h.shape, 1), v.region.mesh.nelements
            )
        else:
            vb = v.region.dhdX

        if u is not None:
            if not grad_u:
                ub = np.tile(
                    u.region.h.reshape(*u.region.h.shape, 1), u.region.mesh.nelements
                )
            else:
                ub = u.region.dhdX

        if u is None:

            if not grad_v:
                return np.einsum("ape,ipe,pe->aie", vb, fun, dV, optimize=True)
            else:
                if parallel:
                    return integrate_gradv(vb, fun, dV)
                else:
                    return np.einsum("aJpe,iJpe,pe->aie", vb, fun, dV, optimize=True)

        else:

            if not grad_v and not grad_u:
                return np.einsum(
                    "ape,...pe,bpe,pe->a...be", vb, fun, ub, dV, optimize=True
                )
            elif grad_v and not grad_u:
                if parallel:
                    return integrate_gradv_u(vb, fun, ub, dV)
                    # return np.einsum("aJpe,iJpe,bpe,pe->aibe", vb, fun, ub, dV)
                else:
                    return np.einsum(
                        "aJpe,iJpe,bpe,pe->aibe", vb, fun, ub, dV, optimize=True
                    )
            elif not grad_v and grad_u:
                if parallel:
                    return integrate_v_gradu(vb, fun, ub, dV)
                    # return np.einsum("ape,kLpe,bLpe,pe->abke", vb, fun, ub, dV)
                else:
                    return np.einsum(
                        "ape,kLpe,bLpe,pe->abke", vb, fun, ub, dV, optimize=True
                    )
            else:  # grad_v and grad_u
                if parallel:
                    return integrate_gradv_gradu(vb, fun, ub, dV)
                else:
                    return np.einsum(
                        "aJpe,iJkLpe,bLpe,pe->aibke", vb, fun, ub, dV, optimize=True
                    )


try:
    from numba import jit, prange

    jitargs = {"nopython": True, "nogil": True, "fastmath": True, "parallel": True}

    @jit(**jitargs)
    def integrate_gradv_u(v, fun, u, dV):

        nnodes_a = v.shape[0]
        nnodes_b = u.shape[0]
        ndim, ngauss, nelems = fun.shape[-3:]

        out = np.zeros((nnodes_a, ndim, nnodes_b, nelems))
        for a in prange(nnodes_a):  # basis function "a"
            for b in prange(nnodes_b):  # basis function "b"
                for p in prange(ngauss):  # integration point "p"
                    for e in prange(nelems):  # element "e"
                        for i in prange(ndim):  # first index "i"
                            for J in prange(ndim):  # second index "J"
                                out[a, i, b, e] += (
                                    v[a, J, p, e]
                                    * u[b, p, e]
                                    * fun[i, J, p, e]
                                    * dV[p, e]
                                )

        return out

    @jit(**jitargs)
    def integrate_v_gradu(v, fun, u, dV):

        nnodes_a = v.shape[0]
        nnodes_b = u.shape[0]
        ndim, ngauss, nelems = fun.shape[-3:]

        out = np.zeros((nnodes_a, nnodes_b, ndim, nelems))
        for a in prange(nnodes_a):  # basis function "a"
            for b in prange(nnodes_b):  # basis function "b"
                for p in prange(ngauss):  # integration point "p"
                    for e in prange(nelems):  # element "e"
                        for k in prange(ndim):  # third index "k"
                            for L in prange(ndim):  # fourth index "L"
                                out[a, b, k, e] += (
                                    v[a, p, e]
                                    * u[b, L, p, e]
                                    * fun[k, L, p, e]
                                    * dV[p, e]
                                )

        return out

    @jit(**jitargs)
    def integrate_gradv(v, fun, dV):

        nnodes = v.shape[0]
        ndim, ngauss, nelems = fun.shape[-3:]

        out = np.zeros((nnodes, ndim, nelems))

        for a in prange(nnodes):  # basis function "a"
            for p in prange(ngauss):  # integration point "p"
                for e in prange(nelems):  # element "e"
                    for i in prange(ndim):  # first index "i"
                        for J in prange(ndim):  # second index "J"
                            out[a, i, e] += v[a, J, p, e] * fun[i, J, p, e] * dV[p, e]

        return out

    @jit(**jitargs)
    def integrate_gradv_gradu(v, fun, u, dV):

        nnodes_a = v.shape[0]
        nnodes_b = u.shape[0]
        ndim, ngauss, nelems = fun.shape[-3:]

        out = np.zeros((nnodes_a, ndim, nnodes_b, ndim, nelems))
        for a in prange(nnodes_a):  # basis function "a"
            for b in prange(nnodes_b):  # basis function "b"
                for p in prange(ngauss):  # integration point "p"
                    for e in prange(nelems):  # element "e"
                        for i in prange(ndim):  # first index "i"
                            for J in prange(ndim):  # second index "J"
                                for k in prange(ndim):  # third index "k"
                                    for L in prange(ndim):  # fourth index "L"
                                        out[a, i, b, k, e] += (
                                            v[a, J, p, e]
                                            * u[b, L, p, e]
                                            * fun[i, J, k, L, p, e]
                                            * dV[p, e]
                                        )

        return out


except:
    pass



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np


def inv(A, detA=None, full_output=False):
    invA = np.zeros_like(A)

    if detA is None:
        detA = det(A)

    if A.shape[0] == 3:

        invA[0, 0] = -A[1, 2] * A[1, 1] + A[1, 1] * A[2, 2]
        invA[1, 1] = -A[0, 2] * A[0, 2] + A[0, 0] * A[2, 2]
        invA[2, 2] = -A[0, 1] * A[0, 1] + A[0, 0] * A[1, 1]

        invA[0, 1] = A[0, 2] * A[1, 2] - A[0, 1] * A[2, 2]
        invA[1, 2] = A[0, 2] * A[0, 1] - A[0, 0] * A[1, 2]
        invA[0, 2] = -A[0, 2] * A[1, 1] + A[0, 1] * A[1, 2]

        invA[1, 0] = invA[0, 1]
        invA[2, 1] = invA[1, 2]
        invA[2, 0] = invA[0, 2]

    elif A.shape[0] == 2:

        invA[0, 0] = A[1, 1]
        invA[0, 1] = -A[0, 1]
        invA[1, 0] = -A[0, 1]
        invA[1, 1] = A[0, 0]

    if full_output:
        return invA / detA, detA
    else:
        return invA / detA


def det(A):

    if A.shape[0] == 3:

        detA = (
            A[0, 0] * A[1, 1] * A[2, 2]
            + A[0, 1] * A[1, 2] * A[0, 2]
            + A[0, 2] * A[0, 1] * A[1, 2]
            - A[0, 2] * A[1, 1] * A[0, 2]
            - A[1, 2] * A[1, 2] * A[0, 0]
            - A[2, 2] * A[0, 1] * A[0, 1]
        )

    elif A.shape[0] == 2:

        detA = A[0, 0] * A[1, 1] - A[0, 1] ** 2

    return detA


def cof(A):
    return inv(A, detA=1.0)


def eigh(A):
    wA, vA = np.linalg.eigh(A.transpose([2, 3, 0, 1]))
    return wA.transpose([2, 0, 1]), vA.transpose([2, 3, 0, 1])


def eigvalsh(A, shear=False):
    wA = np.linalg.eigh(A.transpose([2, 3, 0, 1]))[0].transpose([2, 0, 1])
    if shear:
        ndim = wA.shape[0]
        if ndim == 3:
            ij = [(1, 0), (2, 0), (2, 1)]
        elif ndim == 2:
            ij = [(1, 0)]
        dwA = np.array([wA[i] - wA[j] for i, j in ij])
        return np.vstack((wA, dwA))
    else:
        return wA



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import pytest
import numpy as np
import felupe as fe


def test_math():
    H = (np.random.rand(3, 3, 8, 200) - 0.5) / 10
    F = fe.math.identity(H) + H
    C = fe.math.dot(fe.math.transpose(F), F)
    A = np.random.rand(3, 3, 3, 3, 8, 200)

    fe.math.dot(C, C)
    fe.math.dot(C, A)
    fe.math.dot(A, C)
    fe.math.dot(A, A)

    fe.math.ddot(C, C)
    fe.math.ddot(C, A)
    fe.math.ddot(A, C)
    fe.math.ddot(A, A)

    fe.math.inv(C)
    fe.math.det(C)
    fe.math.cof(C)
    fe.math.dya(C, C)
    fe.math.cdya_ik(F, F)
    fe.math.cdya_il(F, F)
    fe.math.cdya(F, F)

    fe.math.tovoigt(C)
    fe.math.eigvals(C)


if __name__ == "__main__":
    test_math()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from pypardiso import spsolve

# from scipy.sparse.linalg import spsolve

from .math import values


def partition(v, K, dof1, dof0, r=None):

    u = values(v)
    if r is None:
        r1 = None
    else:
        r1 = r[dof1]

    u0 = u.ravel()[dof0]
    K11 = K[dof1, :][:, dof1]
    K10 = K[dof1, :][:, dof0]
    return u, u0, K11, K10, dof1, dof0, r1


def solve(u, u0, K11, K10, dof1, dof0, r1=None, u0ext=None):

    if r1 is None:
        r1 = np.zeros(len(dof1))

    if u0ext is None:
        dr0 = np.zeros(len(dof0))
    else:
        dr0 = K10.dot(u0ext - u0)

    du1 = spsolve(K11, -r1 - dr0.reshape(*r1.shape))
    du = np.empty(u.size)
    du[dof1] = du1
    du[dof0] = u0ext - u0
    return du.reshape(*u.shape)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 29 18:29:15 2021

@author: adutz
"""
# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from copy import copy, deepcopy

from .math import identity


def extract(fields, fieldgrad=(True, False), add_identity=False):
    grads = np.pad(fieldgrad, (0, len(fields)))
    list_of_fields = []
    for grad, field in zip(grads, fields):
        if grad:
            f = field.grad()
            if add_identity:
                f += identity(f)
        else:
            f = field.interpolate()
        list_of_fields.append(f)
    return list_of_fields


class Indices:
    def __init__(self, eai, ai, region, dim):
        self.eai = eai
        self.ai = ai
        self.dof = np.arange(region.nnodes * dim).reshape(-1, dim)
        self.shape = (region.nnodes * dim, 1)


class Field:
    "n-dimensional field in region."

    def __init__(self, region, dim=1, values=0):
        self.region = region
        self.dim = dim

        # init values
        self.values = np.ones((region.nnodes, dim)) * values

        eai, ai = self.indices_per_element(self.region.connectivity, dim)
        self.indices = Indices(eai, ai, region, dim)

    def indices_per_element(self, connectivity, dim):
        "Pre-defined indices for sparse matrices."

        # index of element "e", node "a" and nodal-value component "i"
        eai = np.stack(
            [dim * np.tile(conn, (dim, 1)).T + np.arange(dim) for conn in connectivity]
        )
        # store indices as (rows, cols) (note: sparse-matrices are always 2d)
        ai = (eai.ravel(), np.zeros_like(eai.ravel()))

        return eai, ai

    def grad(self):
        "gradient dudX_IJpe"
        # gradient as partial derivative of nodal field values "aI"
        # w.r.t. undeformed coordiante "J" evaluated at quadrature point "p"
        # for element "e"
        return np.einsum(
            "ea...,aJpe->...Jpe",
            self.values[self.region.connectivity],
            self.region.dhdX,
        )

    def interpolate(self):
        "interpolated values u_Ipe"
        # interpolated field values "aI"
        # evaluated at quadrature point "p"
        # for element "e"
        return np.einsum(
            "ea...,ap->...pe", self.values[self.region.connectivity], self.region.h
        )

    def copy(self):
        out = copy(self)
        return out

    def full(self, a):
        self.values = np.full(self.values.shape, a, dtype=np.float)

    def fill(self, a):
        self.full(a)

    def __add__(self, newvalues):
        self.__iadd__(newvalues)

    def __sub__(self, newvalues):
        self.__isub__(newvalues)

    def __mul__(self, newvalues):
        self.__imul__(newvalues)

    def __truediv__(self, newvalues):
        self.__itruediv__(newvalues)

    def __iadd__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values += newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values += newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __isub__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values -= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values -= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __imul__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values *= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values *= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __itruediv__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values /= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values /= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")
